from decimal import Decimal
from requests_flask_adapter import Session

import pytest

from crudest import create_access_token, create_refresh_token

from .cat_api import app, db


@pytest.fixture
def client():
    # fancier alternative for app.test_client() that uses requests
    app.config['SERVER_NAME'] = 'feline.io'
    Session.register('http://feline.io', app)
    return Session()


@pytest.fixture
def access_token():
    with app.app_context():
        yield create_access_token(identity=1, fresh=True)


@pytest.fixture
def refresh_token():
    with app.app_context():
        yield create_refresh_token(identity=1)


@pytest.fixture
def database():
    db.clear()
    db['User'][1] = {'id': 1, 'name': 'Jon Arbuckle'}
    db['Cat'][1] = {'id': 1, 'name': 'Garfield', 'weight': Decimal('24.67')}
    db['Cat'][2] = {'id': 2, 'name': 'Leftfield', 'weight': Decimal('23.28')}
    db['Cat'][3] = {'id': 3, 'name': 'Rightfield', 'weight': Decimal('22.73')}
    db['CatWhisker'][1] = {'id': 1, 'cat_id': 1, 'length': Decimal('10.57')}
    db['CatWhisker'][2] = {'id': 2, 'cat_id': 1, 'length': Decimal('11.03')}
    db['CatWhisker'][3] = {'id': 3, 'cat_id': 1, 'length': Decimal('9.95')}
    db['CatWhisker'][4] = {'id': 4, 'cat_id': 1, 'length': Decimal('10.34')}
    db['CatSync']['URQpbCZ28urcWnEEeCOh3JAbol0XlAax'] = {'id': 'URQpbCZ28urcWnEEeCOh3JAbol0XlAax', 'done': True}
    return db


def test_create_cat(client, access_token, database):
    name = 'Simba'
    weight = Decimal('12.34')

    rv = client.post('http://feline.io/cats', headers={
        'Authorization': 'Bearer ' + access_token
    }, data={
        'name': name,
        'weight': weight
    })

    assert rv.status_code == 201
    stored = database['Cat'][rv.json()['id']]
    assert stored['name'] == name
    assert stored['weight'] == weight


def test_create_cat_whisker(client, access_token, database):
    cat_id = 1
    length = Decimal('12.34')

    rv = client.post(f'http://feline.io/cats/{cat_id}/whiskers', headers={
        'Authorization': 'Bearer ' + access_token
    }, data={
        'length': length
    })

    assert rv.status_code == 201
    stored = database['CatWhisker'][rv.json()['id']]
    assert stored['cat_id'] == cat_id
    assert stored['length'] == length


def test_create_cat_action(client, access_token, database):
    cat_id = 1
    verb = 'meow'

    rv = client.post(f'http://feline.io/cats/{cat_id}/actions', headers={
        'Authorization': 'Bearer ' + access_token
    }, data={
        'verb': verb
    })

    assert rv.status_code == 201
    assert rv.json()['links']['collection'] == 'http://feline.io/cats/1/actions'


def test_create_cat_sync(client, access_token, database):
    cat_id = 1

    rv = client.post(f'http://feline.io/cats/{cat_id}/syncs', headers={
        'Authorization': 'Bearer ' + access_token
    })

    assert rv.status_code == 201
    assert not rv.json()['done']


def test_list_cats(client, access_token, database):
    rv = client.get('http://feline.io/cats', headers={
        'Authorization': 'Bearer ' + access_token
    })
    results = rv.json()

    assert rv.status_code == 200
    assert isinstance(results, list)
    assert len(results) == 2  # page size is 2

    first_result = results[0]
    stored = database['Cat'][1]
    for key in stored.keys():
        assert str(first_result[key]) == str(stored[key])  # to str because decimals

    assert 'link' in rv.headers
    assert rv.links['self']['url'] == 'http://feline.io/cats'
    assert rv.links['next']['url'] == 'http://feline.io/cats?page=2'

    rv = client.get(f'http://feline.io/cats?page=2', headers={
        'Authorization': 'Bearer ' + access_token
    })
    results = rv.json()

    assert rv.status_code == 200
    assert len(results) == 1  # only 1 left on last page
    assert 'next' not in rv.links


def test_list_cat_whiskers(client, access_token, database):
    rv = client.get('http://feline.io/cats/1/whiskers', headers={
        'Authorization': 'Bearer ' + access_token
    })
    results = rv.json()

    assert rv.status_code == 200
    assert isinstance(results, dict)
    assert 'data' in results
    assert len(results['data']) == 3  # page size is 3

    first_result = results['data'][0]
    stored = database['CatWhisker'][1]
    for key in stored.keys():
        assert str(first_result[key]) == str(stored[key])  # to str because decimals

    assert 'links' in results
    assert results['links']['self'] == 'http://feline.io/cats/1/whiskers'
    assert results['links']['next'] == 'http://feline.io/cats/1/whiskers?page=2'

    rv = client.get(f'http://feline.io/cats/1/whiskers?page=2', headers={
        'Authorization': 'Bearer ' + access_token
    })
    results = rv.json()

    assert rv.status_code == 200
    assert len(results['data']) == 1  # only 1 left on last page
    assert 'next' not in results['links']


def test_retrieve_cat(client, access_token, database):
    rv = client.get('http://feline.io/cats/1', headers={
        'Authorization': 'Bearer ' + access_token
    })
    result = rv.json()

    assert rv.status_code == 200
    stored = database['Cat'][1]
    for key in stored.keys():
        assert str(result[key]) == str(stored[key])  # to str because decimals


def test_retrieve_cat_whisker(client, access_token, database):
    rv = client.get('http://feline.io/cats/1/whiskers/1', headers={
        'Authorization': 'Bearer ' + access_token
    })
    result = rv.json()

    assert rv.status_code == 200
    stored = database['CatWhisker'][1]
    for key in stored.keys():
        assert str(result[key]) == str(stored[key])  # to str because decimals


def test_retrieve_cat_sync(client, access_token, database):
    cat_sync_id = 'URQpbCZ28urcWnEEeCOh3JAbol0XlAax'

    rv = client.get(f'http://feline.io/cats/1/syncs/{cat_sync_id}', headers={
        'Authorization': 'Bearer ' + access_token
    })
    result = rv.json()

    assert rv.status_code == 200
    stored = database['CatSync'][cat_sync_id]
    for key in stored.keys():
        assert result[key] == stored[key]


def test_update_cat(client, access_token, database):
    name = 'Garfield aka The Fat Cat'

    rv = client.put('http://feline.io/cats/1', headers={
        'Authorization': 'Bearer ' + access_token
    }, data={
        'name': name
    })
    result = rv.json()

    assert rv.status_code == 200
    stored = database['Cat'][result['id']]
    assert stored['name'] == name
    assert stored['weight'] == Decimal(result['weight'])


def test_update_cat_whisker(client, access_token, database):
    length = Decimal('9.99')

    rv = client.put('http://feline.io/cats/1/whiskers/1', headers={
        'Authorization': 'Bearer ' + access_token
    }, data={
        'length': length
    })
    result = rv.json()

    assert rv.status_code == 200
    stored = database['CatWhisker'][result['id']]
    assert stored['length'] == Decimal(result['length'])


def test_delete_cat(client, access_token, database):
    rv = client.delete('http://feline.io/cats/1', headers={
        'Authorization': 'Bearer ' + access_token
    })

    assert rv.status_code == 204
    assert 1 not in database['Cat']
    assert 1 not in set(cw['cat_id'] for cw in database['CatWhisker'].values())


def test_delete_cat_whisker(client, access_token, database):
    rv = client.delete('http://feline.io/cats/1/whiskers/1', headers={
        'Authorization': 'Bearer ' + access_token
    })

    assert rv.status_code == 204
    assert 1 not in database['CatWhisker']



import functools
import re
from abc import ABCMeta, abstractmethod

import flask_jwt_extended
from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from flask import jsonify, request
from flask.helpers import make_response, url_for
from flask.views import MethodView
from flask_jwt_extended import (JWTManager, create_access_token,
                                create_refresh_token, get_jti, get_jwt_claims,
                                get_jwt_identity)
from flask_swagger_ui import get_swaggerui_blueprint
from webargs import fields
from webargs.flaskparser import parser, use_kwargs

__all__ = [
    'Resource',
    'CreateResource',
    'NonListableRetrieveResource',
    'RetrieveResource',
    'UpdateResource',
    'DeleteResource',
    'CrudResource',
    'RestApi',
    'Response',
    'HeadedResponse',
    'WrappedResponse',
    'fields',
    'extra_args',
    'create_access_token',
    'create_refresh_token',
    'get_jwt_identity',
    'get_jwt_claims',
    'get_jti',
    'basic_auth_required',
    'jwt_required',
    'jwt_optional',
    'fresh_jwt_required',
    'jwt_refresh_token_required',
]


class Resource(metaclass=ABCMeta):
    pass


class CreateResource(Resource):

    @abstractmethod
    def create(self, *args, **kwargs):
        raise NotImplementedError()


class NonListableRetrieveResource(Resource):

    @abstractmethod
    def retrieve(self, *args, **kwargs):
        raise NotImplementedError()


class RetrieveResource(NonListableRetrieveResource):

    @abstractmethod
    def list(self, *args, **kwargs):
        raise NotImplementedError()


class UpdateResource(Resource):

    @abstractmethod
    def update(self, *args, **kwargs):
        raise NotImplementedError()


class DeleteResource(Resource):

    @abstractmethod
    def delete(self, *args, **kwargs):
        raise NotImplementedError()


class CrudResource(CreateResource, RetrieveResource, UpdateResource, DeleteResource):
    pass


class Response:
    """
    Response objects can be used to pass metadata in your API responses. In 
    particular, they allow the passing of relevant link relations. Some of these
    links get added automatically, but you can also add more yourself. The link
    names should be taken from the standard IANA Link Relation Registry:
    
    https://www.iana.org/assignments/link-relations/link-relations.xhtml
    """

    def __init__(self, data, links=None):
        self._data = data
        self._links = links if links is not None else {}

    def dump_data(self, schema_cls, many=False):
        return schema_cls(many=many).dump(self._data)
    
    def extend_links(self, base_links):
        if base_links is None:
            return self._links
        else:
            return {**base_links, **self._links}

    def generate(self, schema_cls, many, base_links=None):
        """
        The base implementation simply generates JSON, ignoring the links.
        """
        return jsonify(self.dump_data(schema_cls, many=many))


class HeadedResponse(Response):
    """
    Adds the links metadata to the response via the HTTP Link header and allows
    any other given headers to be added as well.
    """

    def __init__(self, data, links=None, headers=None):
        super().__init__(data, links)
        self._headers = headers

    def generate(self, schema_cls, many, base_links=None):
        response = make_response(super().generate(schema_cls, many=many))
        response.headers['Link'] = ', '.join(
            f'<{u}>; rel="{n}"'
            for (n, u) in self.extend_links(base_links).items()
        )
        if self._headers is not None:
            for key, value in self._headers.items():
                response.headers[key] = value
        return response


class WrappedResponse(Response):
    """
    Wraps the actual response content under a top-level 'data' member in the
    generated JSON, so that other metadata can be added to the response under
    different top-level members.
    """

    def __init__(self, data, links=None, **kwargs):
        super().__init__(data, links)
        self._kwargs = kwargs
    
    def generate(self, schema_cls, many, base_links=None):
        return jsonify(
            data=self.dump_data(schema_cls, many=many),
            links=self.extend_links(base_links),
            **self._kwargs
        )


class RestView(MethodView):

    @staticmethod
    def _extract_parent_ids(resource, kwargs):
        return {p: kwargs[p] for p in resource.id_params[:-1]}

    def __init__(self, schema_cls, resource, num_ids=1):
        super().__init__()
        self.schema_cls = schema_cls
        self.resource = resource
        self.num_ids = num_ids

    def post(self, **kwargs):
        parent_ids = self._extract_parent_ids(self.resource, kwargs)
        kwargs.update(parser.parse(self.schema_cls(), request))
        response = self.resource.create(**kwargs)
        if not isinstance(response, Response):
            response = Response(data=response)
        return response.generate(self.schema_cls, many=False, base_links={
            'collection': url_for(self.resource.name, _external=True, **parent_ids)
        }), 201

    def get(self, **kwargs):
        parent_ids = self._extract_parent_ids(self.resource, kwargs)
        if len(kwargs) < self.num_ids:
            response = self.resource.list(**kwargs)
            if not isinstance(response, Response):
                response = Response(data=response)
            return response.generate(self.schema_cls, many=True, base_links={
                'self': url_for(self.resource.name, _external=True, **parent_ids)
            })
        else:
            response = self.resource.retrieve(**kwargs)
            if not isinstance(response, Response):
                response = Response(data=response)
            return response.generate(self.schema_cls, many=False, base_links={
                'self': url_for(
                    self.resource.name,
                    _external=True,
                    **{**parent_ids, **kwargs}
                ),
                'collection': url_for(
                    self.resource.name,
                    _external=True,
                    **parent_ids
                )
            })

    def put(self, **kwargs):
        parent_ids = self._extract_parent_ids(self.resource, kwargs)
        kwargs.update(parser.parse(self.schema_cls(partial=True), request))
        response = self.resource.update(**kwargs)
        if not isinstance(response, Response):
            response = Response(data=response)
        return response.generate(self.schema_cls, many=False, base_links={
            'collection': url_for(self.resource.name, _external=True, **parent_ids)
        })

    def delete(self, **kwargs):
        self.resource.delete(**kwargs)
        return '', 204


class RestApi:

    RE_URL = re.compile(r'<(?:[^:<>]+:)?([^<>]+)>')

    def __init__(self, app, title, version='v1', spec_path='/spec', docs_path='/docs'):
        self.app = app
        self.resource_methods = {}

        marshmallow_plugin = MarshmallowPlugin()
        self.spec = spec = APISpec(
            title=title,
            version=version,
            openapi_version="3.0.2",
            plugins=[marshmallow_plugin],
        )
        self.openapi = marshmallow_plugin.converter  # openapi converter from marshmallow plugin

        self.jwt = JWTManager(app)
        self.token_in_blacklist_checker = self.jwt.token_in_blacklist_loader  # shortcut to blacklisting decorator
        self.token_claims_loader = self.jwt.user_claims_loader  # shortcut to user claims decorator

        spec.components.security_scheme('basic_http', {
            'type': 'http',
            'scheme': 'basic'
        })
        spec.components.security_scheme('jwt_access_token', {
            'type': 'http',
            'scheme': 'bearer',
            'bearerFormat': 'JWT'
        })
        spec.components.security_scheme('jwt_refresh_token', {
            'type': 'http',
            'scheme': 'bearer',
            'bearerFormat': 'JWT'
        })

        @app.route(spec_path)
        def get_spec():
            return jsonify(spec.to_dict())

        swaggerui_blueprint = get_swaggerui_blueprint(docs_path, spec_path)
        app.register_blueprint(swaggerui_blueprint, url_prefix=docs_path)

    def resource(self, path, name, schema):
        self.spec.components.schema(name, schema=schema)

        def decorator(cls):
            cls.name = name
            cls.id_params = self.RE_URL.findall(path)

            base_path = '/'.join(path.split('/')[:-1])
            view = RestView.as_view(name, schema, cls(), len(cls.id_params))
            if issubclass(cls, CreateResource):
                self.add_path(base_path, view, method='POST',
                              tag=name, id_params=cls.id_params[:-1],
                              input_schema=schema, output_schema=schema,
                              extra_args=getattr(cls.create, '__extra_args__', None),
                              auth_required=getattr(cls.create, '__auth_required__', None),
                              status_code=201, description=cls.create.__doc__)
            if issubclass(cls, RetrieveResource):
                self.add_path(base_path, view, method='GET',
                              tag=name, id_params=cls.id_params[:-1],
                              output_schema=schema(many=True),
                              extra_args=getattr(cls.list, '__extra_args__', None),
                              auth_required=getattr(cls.list, '__auth_required__', None),
                              description=cls.list.__doc__)
            if issubclass(cls, NonListableRetrieveResource):
                self.add_path(path, view, method='GET',
                              tag=name, id_params=cls.id_params,
                              output_schema=schema,
                              extra_args=getattr(cls.retrieve, '__extra_args__', None),
                              auth_required=getattr(cls.retrieve, '__auth_required__', None),
                              description=cls.retrieve.__doc__)
            if issubclass(cls, UpdateResource):
                self.add_path(path, view, method='PUT',
                              tag=name, id_params=cls.id_params,
                              input_schema=schema, output_schema=schema,
                              extra_args=getattr(cls.update, '__extra_args__', None),
                              auth_required=getattr(cls.update, '__auth_required__', None),
                              description=cls.update.__doc__)
                self.app.add_url_rule(path, view_func=view, methods=['PUT'])
            if issubclass(cls, DeleteResource):
                self.add_path(path, view, method='DELETE',
                              tag=name, id_params=cls.id_params,
                              extra_args=getattr(cls.delete, '__extra_args__', None),
                              auth_required=getattr(cls.delete, '__auth_required__', None),
                              status_code=204, description=cls.delete.__doc__)

            # keep track of methods
            self.resource_methods[name] = next(
                r.methods for r in self.app.url_map.iter_rules() if r.endpoint == name
            )
                            
            return cls

        return decorator

    def add_path(self, path, view, method, tag, id_params=None,
                 input_schema=None, output_schema=None, extra_args=None, auth_required=None,
                 status_code='default', description=''):
        swagger_path = self.RE_URL.sub(r'{\1}', path)
        self.app.add_url_rule(path, view_func=view, methods=[method])

        parameters = [{'name': id_param, 'in': 'path'} for id_param in id_params]
        if extra_args:
            parameters.extend(self.openapi.fields2parameters(extra_args, default_in='query'))

        request_body = {}
        if input_schema:
            request_body['requestBody'] = {
                'content': {
                    'application/json': {
                        'schema': self.openapi.schema2jsonschema(input_schema)
                    }
                }
            }

        self.spec.path(swagger_path, operations={
            method.lower(): {
                'description': description,
                'parameters': parameters,
                'responses': {
                    status_code: {'schema': output_schema} if output_schema else {}
                },
                'tags': [tag],
                'security': [{auth_required: []}] if auth_required else [],
                **request_body
            }
        })

    def url_for(self, resource_name, _method=None, _external=True, **kwargs):
        if _method is None:
            methods = self.resource_methods[resource_name]
            for desired_method in ('GET', 'PUT', 'DELETE', 'POST'):
                if desired_method in methods:
                    _method = desired_method
                    break
        return url_for(resource_name, _method=_method, _external=_external, **kwargs)


def extra_args(args):

    def decorator(func):
        if getattr(func, '__extra_args__', None) is None:
            func.__extra_args__ = {}
        func.__extra_args__.update(args)
        return use_kwargs(args)(func)

    return decorator


def basic_auth_required(func):
    func.__auth_required__ = 'basic_http'
    return func


def jwt_required(func):
    func.__auth_required__ = 'jwt_access_token'
    return flask_jwt_extended.jwt_required(func)


def jwt_optional(func):
    func.__auth_required__ = 'jwt_access_token'
    return flask_jwt_extended.jwt_optional(func)


def fresh_jwt_required(func):
    func.__auth_required__ = 'jwt_access_token'
    return flask_jwt_extended.fresh_jwt_required(func)


def jwt_refresh_token_required(func):
    func.__auth_required__ = 'jwt_refresh_token'
    return flask_jwt_extended.jwt_refresh_token_required(func)



from setuptools import setup

setup(
    name = 'crudest',
    version = '0.6',
    description = 'CRUD-structured REST APIs in Python, using Flask, marshmallow, webargs and apispec',
    license = 'Apache Software License (ASF)',
    url = 'https://github.com/trackuity/crudest',
    py_modules = ['crudest'],
    install_requires = ['Flask', 'apispec>3.0', 'webargs', 'marshmallow>3.0',
                        'flask-swagger-ui>3.0', 'Flask-JWT-Extended']
)



import random
import string
from typing import collections

from flask.app import Flask
from flask.json import jsonify
from marshmallow import Schema, fields

from crudest import (CreateResource, CrudResource, DeleteResource, HeadedResponse, NonListableRetrieveResource, RestApi,
                     RetrieveResource, UpdateResource, WrappedResponse, extra_args, jwt_required)

app = Flask(__name__)
app.config['JWT_SECRET_KEY'] = 's3cr1t'

api = RestApi(app, title='Cat API', version='v1')

db = collections.defaultdict(dict)


class InvalidUsage(Exception):
    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['msg'] = self.message
        return rv


@app.errorhandler(InvalidUsage)
def handle_invalid_usage(error):
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    return response


class IdSchema(Schema):
    id = fields.Int()


class CatSchema(IdSchema):
    name = fields.Str()
    weight = fields.Decimal(as_string=True)


class CatWhiskerSchema(IdSchema):
    cat_id = fields.Int()
    length = fields.Decimal(as_string=True)


class CatActionSchema(IdSchema):
    verb = fields.Str()


class CatSyncSchema(Schema):
    id = fields.Str()  # str, not int
    done = fields.Bool()


@api.resource('/cats/<int:cat_id>', name='Cat', schema=CatSchema)
class CatResource(CreateResource, RetrieveResource, UpdateResource, DeleteResource):

    @jwt_required
    def create(self, **kwargs):
        cat_id = max(db['Cat'].keys() or [0]) + 1
        cat = {'id': cat_id, **kwargs}
        db['Cat'][cat_id] = cat
        return cat

    @jwt_required
    @extra_args({'page': fields.Int(missing=1)})
    def list(self, page):
        links = {}
        values = list(db['Cat'].values())
        if (page + 1) * 2 <= (len(values) // 2 + 1) * 2:  # is there another page?
            links['next'] = api.url_for('Cat', page=page+1)
        # grabbing this opportunity to test headed responses!
        return HeadedResponse(
            data=values[(page-1)*2:page*2],
            links=links
        )

    @jwt_required
    def retrieve(self, cat_id):
        if cat_id not in db['Cat']:
            raise InvalidUsage('Cat not found.', status_code=404)
        return db['Cat'][cat_id]

    @jwt_required
    def update(self, cat_id, **kwargs):
        if cat_id not in db['Cat']:
            raise InvalidUsage('Cat not found.', status_code=404)
        updated_cat = {**db['Cat'][cat_id], **kwargs}
        db['Cat'][cat_id] = updated_cat
        return updated_cat

    @jwt_required
    def delete(self, cat_id):
        if cat_id not in db['Cat']:
            raise InvalidUsage('Cat not found.', status_code=404)
        del db['Cat'][cat_id]
        # delete corresponding whiskers as well
        for cat_whisker in list(db['CatWhisker'].values()):
            if cat_whisker['cat_id'] == cat_id:
                del db['CatWhisker'][cat_whisker['id']]


@api.resource('/cats/<int:cat_id>/whiskers/<int:cat_whisker_id>', name='CatWhisker', schema=CatWhiskerSchema)
class CatWhiskerResource(CrudResource):

    @jwt_required
    def create(self, cat_id, **kwargs):
        if cat_id not in db['Cat']:
            raise InvalidUsage('Cat not found.', status_code=404)
        cat_whisker_id = max(db['CatWhisker'].keys() or [0]) + 1
        cat_whisker = {'id': cat_whisker_id, 'cat_id': cat_id, **kwargs}
        db['CatWhisker'][cat_whisker_id] = cat_whisker
        return cat_whisker

    @jwt_required
    @extra_args({'page': fields.Int(missing=1)})
    def list(self, cat_id, page):
        if cat_id not in db['Cat']:
            raise InvalidUsage('Cat not found.', status_code=404)
        links = {}
        values = [
            value for value in db['CatWhisker'].values()
            if value['cat_id'] == cat_id
        ]
        if (page + 1) * 3 <= (len(values) // 3 + 1) * 3:  # is there another page?
            links['next'] = api.url_for('CatWhisker', cat_id=cat_id, page=page+1)
        # grabbing this opportunity to test wrapped responses!
        return WrappedResponse(
            data=values[(page-1)*3:page*3],
            links=links
        )

    @jwt_required
    def retrieve(self, cat_id, cat_whisker_id):
        if cat_id not in db['Cat']:
            raise InvalidUsage('Cat not found.', status_code=404)
        if cat_whisker_id not in db['CatWhisker']:
            raise InvalidUsage('Whisker not found.', status_code=404)
        return db['CatWhisker'][cat_whisker_id]

    @jwt_required
    def update(self, cat_id, cat_whisker_id, **kwargs):
        if cat_id not in db['Cat']:
            raise InvalidUsage('Cat not found.', status_code=404)
        if cat_whisker_id not in db['CatWhisker']:
            raise InvalidUsage('Whisker not found.', status_code=404)
        updated_cat_whisker = {**db['CatWhisker'][cat_whisker_id], **kwargs}
        db['CatWhisker'][cat_whisker_id] = updated_cat_whisker
        return updated_cat_whisker

    @jwt_required
    def delete(self, cat_id, cat_whisker_id):
        if cat_id not in db['Cat']:
            raise InvalidUsage('Cat not found.', status_code=404)
        if cat_whisker_id not in db['CatWhisker']:
            raise InvalidUsage('Whisker not found.', status_code=404)
        del db['CatWhisker'][cat_whisker_id]


@api.resource('/cats/<int:cat_id>/actions/<int:cat_action_id>', name='CatAction', schema=CatActionSchema)
class CatActionResource(CreateResource, UpdateResource):

    @jwt_required
    def create(self, cat_id, verb):
        return self.update(cat_id, 1, verb)

    @jwt_required
    def update(self, cat_id, cat_action_id, verb):
        if cat_id not in db['Cat']:
            raise InvalidUsage('Cat not found.', status_code=404)
        cat_name = db['Cat'][cat_id]['name']
        return WrappedResponse(
            data={
                'id': 1,
                'result': f'The cat named {cat_name} {verb}ed.'
            },
            links=dict(
                self=api.url_for('CatAction', cat_id=cat_id, cat_action_id=1)
            )
        )


@api.resource('/cats/<int:cat_id>/syncs/<cat_sync_id>', name='CatSync', schema=CatSyncSchema)
class CatSyncResource(CreateResource, NonListableRetrieveResource):

    @jwt_required
    def create(self, cat_id):
        if cat_id not in db['Cat']:
            raise InvalidUsage('Cat not found.', status_code=404)
        cat_sync_id = ''.join([random.choice(string.ascii_letters + string.digits) for n in range(32)])
        cat_sync = {'id': cat_sync_id, 'done': False}
        db['CatSync'][cat_sync_id] = cat_sync
        return cat_sync

    @jwt_required
    def retrieve(self, cat_id, cat_sync_id):
        if cat_id not in db['Cat']:
            raise InvalidUsage('Cat not found.', status_code=404)
        cat_sync = db['CatSync'].get(cat_sync_id)
        if cat_sync is None:
            raise InvalidUsage('Cat sync not found.', status_code=404)
        return cat_sync
